# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

###############################################################################
#
#
# This is a port of the code by Krzysztof Choromanski, Deepali Jain and Vikas
# Sindhwani, based on the portfolio of Blackbox optimization algorithms listed
# below:
#
# "On Blackbox Backpropagation and Jacobian Sensing"; K. Choromanski,
#  V. Sindhwani, NeurIPS 2017
# "Optimizing Simulations with Noise-Tolerant Structured Exploration"; K.
#  Choromanski, A. Iscen, V. Sindhwani, J. Tan, E. Coumans, ICRA 2018
# "Structured Evolution with Compact Architectures for Scalable Policy
#  Optimization"; K. Choromanski, M. Rowland, V. Sindhwani, R. Turner, A.
#  Weller, ICML 2018, https://arxiv.org/abs/1804.02395
#  "From Complexity to Simplicity: Adaptive ES-Active Subspaces for Blackbox
#   Optimization";  K. Choromanski, A. Pacchiano, J. Parker-Holder, Y. Tang, V.
#   Sindhwani, NeurIPS 2019
# "i-Sim2Real: Reinforcement Learning on Robotic Policies in Tight Human-Robot
#  Interaction Loops"; L. Graesser, D. D'Ambrosio, A. Singh, A. Bewley, D. Jain,
#  K. Choromanski, P. Sanketi , CoRL 2022, https://arxiv.org/abs/2207.06572
# "Agile Catching with Whole-Body MPC and Blackbox Policy Learning"; S.
#  Abeyruwan, A. Bewley, N. Boffi, K. Choromanski, D. D'Ambrosio, D. Jain, P.
#  Sanketi, A. Shankar, V. Sindhwani, S. Singh, J. Slotine, S. Tu, L4DC,
#  https://arxiv.org/abs/2306.08205
# "Robotic Table Tennis: A Case Study into a High Speed Learning System"; A.
#  Bewley, A. Shankar, A. Iscen, A. Singh, C. Lynch, D. D'Ambrosio, D. Jain,
#  E. Coumans, G. Versom, G. Kouretas, J. Abelian, J. Boyd, K. Oslund,
#  K. Reymann, K. Choromanski, L. Graesser, M. Ahn, N. Jaitly, N. Lazic,
#  P. Sanketi, P. Xu, P. Sermanet, R. Mahjourian, S. Abeyruwan, S. Kataoka,
#  S. Moore, T. Nguyen, T. Ding, V. Sindhwani, V. Vanhoucke, W. Gao, Y. Kuang,
#  to be presented at RSS 2023
###############################################################################
r"""Library of gradient ascent algorithms.

Library of stateful gradient ascent algorithms taking as input the gradient and
current parameters, and output the new parameters.
"""

import abc

import numpy as np
import numpy.typing as npt
from typing import List, Optional


class GradientAscentOptimizer(metaclass=abc.ABCMeta):
  """Abstract class for general gradient ascent optimizers.

  Class is responsible for encoding different gradient ascent optimization
  techniques.
  """

  @abc.abstractmethod
  def run_step(self, current_input: npt.NDArray[np.float32],
               gradient: npt.NDArray[np.float32]) -> npt.NDArray[np.float32]:
    """Conducts a single step of gradient ascent optimization.

    Conduct a single step of gradient ascent optimization procedure, given the
    current parameters and the raw gradient.

    Args:
      current_input: the current parameters.
      gradient: the raw gradient.

    Returns:
      New parameters by conducting a single step of gradient ascent.
    """
    raise NotImplementedError("Abstract method")

  @abc.abstractmethod
  def get_state(self) -> List[float]:
    """Returns the state of the optimizer.

    Returns the state of the optimizer.

    Args:

    Returns:
      The state of the optimizer.
    """
    raise NotImplementedError("Abstract method")

  @abc.abstractmethod
  def set_state(self, state: npt.NDArray[np.float32]) -> None:
    """Sets up the internal state of the optimizer.

    Sets up the internal state of the optimizer.

    Args:
      state: state to be set up

    Returns:
    """
    raise NotImplementedError("Abstract method")


class MomentumOptimizer(GradientAscentOptimizer):
  """Class implementing momentum gradient ascent optimizer.

  Setting momentum coefficient to zero is equivalent to vanilla gradient
  ascent.

  the state is the moving average as a list
  """

  def __init__(self, step_size: float, momentum: float):
    self.step_size = step_size
    self.momentum = momentum

    self.moving_average = np.asarray([], dtype=np.float32)
    super().__init__()

  def run_step(self, current_input: npt.NDArray[np.float32],
               gradient: npt.NDArray[np.float32]) -> npt.NDArray[np.float32]:
    if self.moving_average.size == 0:
      # Initialize the moving average
      self.moving_average = np.zeros(len(current_input), dtype=np.float32)
    elif len(self.moving_average) != len(current_input):
      raise ValueError(
          "Dimensions of the parameters and moving average do not match")

    self.moving_average = self.momentum * self.moving_average + (
        1 - self.momentum) * gradient
    step = self.step_size * self.moving_average

    return current_input + step

  def get_state(self) -> List[float]:
    return self.moving_average.tolist()

  def set_state(self, state: npt.NDArray[np.float32]) -> None:
    self.moving_average = np.asarray(state, dtype=np.float32)


class AdamOptimizer(GradientAscentOptimizer):
  """Class implementing ADAM gradient ascent optimizer.

  The state is the first moment moving average, the second
  moment moving average, and t (current step number)
  combined in that order into one list
  """

  def __init__(self,
               step_size: float,
               beta1: Optional[float] = 0.9,
               beta2: Optional[float] = 0.999,
               epsilon: Optional[float] = 1e-07):
    self.step_size = step_size
    self.beta1 = beta1
    self.beta2 = beta2
    self.epsilon = epsilon

    self.first_moment_moving_average = np.asarray([], dtype=np.float32)
    self.second_moment_moving_average = np.asarray([], dtype=np.float32)
    self.t = 0
    super().__init__()

  def run_step(self, current_input: npt.NDArray[np.float32],
               gradient: npt.NDArray[np.float32]) -> npt.NDArray[np.float32]:
    if self.first_moment_moving_average.size == 0:
      # Initialize the moving averages
      self.first_moment_moving_average = np.zeros(
          len(current_input), dtype=np.float32)
      self.second_moment_moving_average = np.zeros(
          len(current_input), dtype=np.float32)
      # Initialize the step counter
      self.t = 0
    elif len(self.first_moment_moving_average) != len(current_input):
      raise ValueError(
          "Dimensions of the parameters and moving averages do not match")

    self.first_moment_moving_average = (
        self.beta1 * self.first_moment_moving_average +
        (1 - self.beta1) * gradient)
    self.second_moment_moving_average = (
        self.beta2 * self.second_moment_moving_average + (1 - self.beta2) *
        (gradient * gradient))

    self.t += 1
    scale = np.sqrt(1 - self.beta2**self.t) / (1 - self.beta1**self.t)

    step = self.step_size * scale * self.first_moment_moving_average / (
        np.sqrt(self.second_moment_moving_average) + self.epsilon)

    return current_input + step

  def get_state(self) -> List[float]:
    return (self.first_moment_moving_average.tolist() +
            self.second_moment_moving_average.tolist() + [self.t])

  def set_state(self, state: npt.NDArray[np.float32]) -> None:
    total_len = len(state)
    if total_len % 2 != 1:
      raise ValueError("The dimension of the state should be odd")
    dim = total_len // 2

    self.first_moment_moving_average = np.asarray(state[:dim], dtype=np.float32)
    self.second_moment_moving_average = np.asarray(
        state[dim:2 * dim], dtype=np.float32)
    self.t = int(state[-1])
    if self.t < 0:
      raise ValueError("The step counter should be non-negative")



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Local Process Pool - based middleware implementation.

This is a simple implementation of a worker pool, running on the local machine.
Each worker object is hosted by a separate process. Each worker object may
handle a number of concurrent requests. The client is given a stub object that
exposes the same methods as the worker, just that they return Futures.

There is a bidirectional pipe between a stub and its corresponding
process/worker. One direction is used to place tasks (method calls), the other
to place results. Tasks and results are correlated by a monotonically
incrementing counter maintained by the stub.

The worker process dequeues tasks promptly and either re-enqueues them to a
local thread pool, or, if the task is 'urgent', it executes it promptly.
"""
import concurrent.futures
import cloudpickle
import dataclasses
import functools
import multiprocessing
import psutil
import threading

from absl import logging
# pylint: disable=unused-import
from compiler_opt.distributed import worker

from contextlib import AbstractContextManager
from multiprocessing import connection
from typing import Any, Callable, Dict, List, Optional


@dataclasses.dataclass(frozen=True)
class Task:
  msgid: int
  func_name: str
  args: tuple
  kwargs: dict
  is_urgent: bool


@dataclasses.dataclass(frozen=True)
class TaskResult:
  msgid: int
  success: bool
  value: Any


def _get_context():
  return multiprocessing.get_context('spawn')


SerializedClass = bytes


def _run_impl(pipe: connection.Connection, worker_class: SerializedClass, *args,
              **kwargs):
  """Worker process entrypoint."""

  # A setting of 1 does not inhibit the while loop below from running since
  # that runs on the main thread of the process. Urgent tasks will still
  # process near-immediately. `threads` only controls how many threads are
  # spawned at a time which execute given tasks. In the typical clang-spawning
  # jobs, this effectively limits the number of clang instances spawned.
  pool = concurrent.futures.ThreadPoolExecutor(max_workers=1)
  obj = cloudpickle.loads(worker_class)(*args, **kwargs)

  # Pipes are not thread safe
  pipe_lock = threading.Lock()

  def send(task_result: TaskResult):
    with pipe_lock:
      pipe.send(task_result)

  def make_ondone(msgid):

    def on_done(f: concurrent.futures.Future):
      if f.exception():
        send(TaskResult(msgid=msgid, success=False, value=f.exception()))
      else:
        send(TaskResult(msgid=msgid, success=True, value=f.result()))

    return on_done

  # Run forever. The stub will just kill the runner when done.
  while True:
    task: Task = pipe.recv()
    the_func = getattr(obj, task.func_name)
    application = functools.partial(the_func, *task.args, **task.kwargs)
    if task.is_urgent:
      try:
        res = application()
        send(TaskResult(msgid=task.msgid, success=True, value=res))
      except BaseException as e:  # pylint: disable=broad-except
        send(TaskResult(msgid=task.msgid, success=False, value=e))
    else:
      pool.submit(application).add_done_callback(make_ondone(task.msgid))


def _run(pipe: connection.Connection, worker_class: SerializedClass, *args,
         **kwargs):
  try:
    _run_impl(pipe, worker_class, *args, **kwargs)
  except BaseException as e:
    logging.error(e)
    raise e


def _make_stub(cls: 'type[worker.Worker]', *args, **kwargs):

  class _Stub:
    """Client stub to a worker hosted by a process."""

    def __init__(self):
      parent_pipe, child_pipe = _get_context().Pipe()
      self._pipe = parent_pipe
      self._pipe_lock = threading.Lock()

      # this is the process hosting one worker instance.
      # we set aside 1 thread to coordinate running jobs, and the main thread
      # to handle high priority requests. The expectation is that the user
      # achieves concurrency through multiprocessing, not multithreading.
      self._process = _get_context().Process(
          target=functools.partial(_run, child_pipe, cloudpickle.dumps(cls), *
                                   args, **kwargs))
      # lock for the msgid -> reply future map. The map will be set to None
      # when we stop.
      self._lock = threading.Lock()
      self._map: Dict[int, concurrent.futures.Future] = {}

      # thread draining the pipe
      self._pump = threading.Thread(target=self._msg_pump)

      # Set the state of this worker to "dead" if the process dies naturally.
      def observer():
        self._process.join()
        # Feed the parent pipe a poison pill, this kills msg_pump
        child_pipe.send(None)

      self._observer = threading.Thread(target=observer)

      # atomic control to _msgid
      self._msgidlock = threading.Lock()
      self._msgid = 0

      # start the worker and the message pump
      self._process.start()
      # the observer must follow the process start, otherwise join() raises.
      self._observer.start()
      self._pump.start()

    def _msg_pump(self):
      while True:
        task_result: Optional[TaskResult] = self._pipe.recv()
        if task_result is None:  # Poison pill fed by observer
          break
        with self._lock:
          future = self._map[task_result.msgid]
          del self._map[task_result.msgid]
        # The following will trigger any callbacks defined on the future, as a
        # direct function call. If those callbacks were set by the scheduler,
        # it's important that self._lock isn't being held when they are being
        # called, otherwise a deadlock could arise from __get_attr__ trying to
        # acquire the lock.
        if task_result.success:
          future.set_result(task_result.value)
        else:
          future.set_exception(task_result.value)

      # clear out pending futures and mark ourselves as "stopped" by null-ing
      # the map
      with self._lock:
        for _, v in self._map.items():
          v.set_exception(concurrent.futures.CancelledError())
        self._map = None

    def _is_stopped(self):
      return self._map is None

    def __getattr__(self, name) -> Callable[[Any], concurrent.futures.Future]:
      result_future = concurrent.futures.Future()

      with self._msgidlock:
        msgid = self._msgid
        self._msgid += 1

      def remote_call(*args, **kwargs):
        with self._lock:
          if self._is_stopped():
            result_future.set_exception(concurrent.futures.CancelledError())
          else:
            with self._pipe_lock:
              self._pipe.send(
                  Task(
                      msgid=msgid,
                      func_name=name,
                      args=args,
                      kwargs=kwargs,
                      is_urgent=cls.is_priority_method(name)))
            self._map[msgid] = result_future
        return result_future

      return remote_call

    def shutdown(self):
      try:
        # Killing the process triggers observer exit, which triggers msg_pump
        # exit
        self._process.kill()
      except:  # pylint: disable=bare-except
        pass

    def set_nice(self, val: int):
      """Sets the nice-ness of the process, this modifies how the OS
      schedules it. Only works on Unix, since val is presumed to be an int.
      """
      psutil.Process(self._process.pid).nice(val)

    def set_affinity(self, val: List[int]):
      """Sets the CPU affinity of the process, this modifies which cores the OS
      schedules it on.
      """
      psutil.Process(self._process.pid).cpu_affinity(val)

    def join(self):
      self._observer.join()
      self._pump.join()
      self._process.join()

    def __dir__(self):
      return [n for n in dir(cls) if not n.startswith('_')]

  return _Stub()


def create_local_worker_pool(worker_cls: 'type[worker.Worker]',
                             count: Optional[int], *args,
                             **kwargs) -> worker.FixedWorkerPool:
  """Create a local worker pool for worker_cls."""
  if not count:
    count = _get_context().cpu_count()
  final_kwargs = worker.get_full_worker_args(worker_cls, **kwargs)
  stubs = [_make_stub(worker_cls, *args, **final_kwargs) for _ in range(count)]
  return worker.FixedWorkerPool(workers=stubs, worker_concurrency=16)


def close_local_worker_pool(pool: worker.FixedWorkerPool):
  """Close the given LocalWorkerPool."""
  # first, trigger killing the worker process and exiting of the msg pump,
  # which will also clear out any pending futures.
  for stub in pool.get_currently_active():
    stub.shutdown()
  # now wait for the message pumps to indicate they exit.
  for stub in pool.get_currently_active():
    stub.join()


class LocalWorkerPoolManager(AbstractContextManager):
  """A pool of workers hosted on the local machines, each in its own process."""

  def __init__(self, worker_class: 'type[worker.Worker]', count: Optional[int],
               *args, **kwargs):
    self._pool = create_local_worker_pool(worker_class, count, *args, **kwargs)

  def __enter__(self) -> worker.FixedWorkerPool:
    return self._pool

  def __exit__(self, *args):
    close_local_worker_pool(self._pool)

  def __del__(self):
    self.__exit__()

  @property
  def stubs(self):
    # Return a shallow copy, to avoid something messing the internal list up
    return list(self._stubs)



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test for worker."""

import gin

from absl.testing import absltest
from compiler_opt.distributed import worker


@gin.configurable(module='_test')
class SomeType:

  def __init__(self, argument):
    pass


class WorkerTest(absltest.TestCase):

  def test_gin_args(self):
    with gin.unlock_config():
      gin.bind_parameter('_test.SomeType.argument', 42)
    real_args = worker.get_full_worker_args(
        SomeType, more_args=2, even_more_args='hi')
    self.assertDictEqual(real_args,
                         dict(argument=42, more_args=2, even_more_args='hi'))


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Worker for ES Training."""

import gin
from typing import List
from compiler_opt.distributed import worker
from compiler_opt.rl import corpus


@gin.configurable
class ESWorker(worker.Worker):
  """Temporary placeholder worker.
  Each time a worker is called, the function value
  it will return increases."""

  def __init__(self, arg, *, kwarg):
    self._arg = arg
    self._kwarg = kwarg
    self.function_value = 0.0

  def temp_compile(self, policy: bytes,
                   samples: List[corpus.ModuleSpec]) -> float:
    if policy and samples:
      self.function_value += 1.0
      return self.function_value
    else:
      return 0.0



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Generate initial training data from the behavior of the current heuristic."""

import concurrent.futures
import contextlib
import functools
import re
from typing import Dict, List, Optional, Union, Tuple  # pylint:disable=unused-import

from absl import app
from absl import flags
from absl import logging
import gin

import tensorflow as tf

from compiler_opt.distributed import worker
from compiler_opt.distributed import buffered_scheduler
from compiler_opt.distributed.local import local_worker_manager

from compiler_opt.rl import compilation_runner
from compiler_opt.rl import corpus
from compiler_opt.rl import policy_saver
from compiler_opt.rl import registry

from tf_agents.system import system_multiprocessing as multiprocessing

# see https://bugs.python.org/issue33315 - we do need these types, but must
# currently use them as string annotations

_DATA_PATH = flags.DEFINE_string('data_path', None,
                                 'Path to folder containing IR files.')
_POLICY_PATH = flags.DEFINE_string(
    'policy_path', '', 'Path to the policy to generate trace with.')
_OUTPUT_PATH = flags.DEFINE_string(
    'output_path', None, 'Path to the output tfrecord file if not None.')
_OUTPUT_PERFORMANCE_PATH = flags.DEFINE_string(
    'output_performance_path', None,
    'Path to the output performance file if not None.')
_NUM_WORKERS = flags.DEFINE_integer(
    'num_workers', None,
    'Number of parallel workers for compilation. `None` for maximum available.')
_SAMPLING_RATE = flags.DEFINE_float(
    'sampling_rate', 1,
    'Sampling rate of modules, 0.5 means 50% sampling rate that generates data '
    'for half modules.')
_MODULE_FILTER = flags.DEFINE_string(
    'module_filter', None,
    'Regex for module names to include, do not provide one to include all')
_KEY_FILTER = flags.DEFINE_string(
    'key_filter', None,
    'Regex for key names to include, do not provide one to include all')
_GIN_FILES = flags.DEFINE_multi_string(
    'gin_files', [], 'List of paths to gin configuration files.')
_GIN_BINDINGS = flags.DEFINE_multi_string(
    'gin_bindings', [],
    'Gin bindings to override the values set in the config files.')

ResultsQueueEntry = Union[Optional[Tuple[str, List[str],
                                         Dict[str,
                                              compilation_runner.RewardStat]]],
                          BaseException]


class FilteringWorker(worker.Worker):
  """Worker that performs a computation and optionally filters the result.


  Args:
    policy_path: the policy_path to generate trace with.
    key_filter: regex filter for key names to include, or None to include all.
  """

  def __init__(self, policy_path: Optional[str], key_filter: Optional[str],
               runner_type: 'type[compilation_runner.CompilationRunner]',
               runner_kwargs):
    self._policy_path = policy_path
    self._key_filter = re.compile(key_filter) if key_filter else None
    self._runner = runner_type(**runner_kwargs)
    self._policy = policy_saver.Policy.from_filesystem(
        policy_path) if policy_path else None

  def compile_and_filter(
      self, loaded_module_spec: corpus.LoadedModuleSpec
  ) -> Tuple[str, List[str], Dict[str, compilation_runner.RewardStat]]:
    data = self._runner.collect_data(
        loaded_module_spec=loaded_module_spec,
        policy=self._policy,
        reward_stat=None,
        model_id=0)
    if self._key_filter is None:
      return (loaded_module_spec.name, data.serialized_sequence_examples,
              data.reward_stats)
    new_reward_stats = {}
    new_sequence_examples = []
    for k, sequence_example in zip(data.keys,
                                   data.serialized_sequence_examples):
      if not self._key_filter.match(k):
        continue
      new_reward_stats[k] = data.reward_stats[k]
      new_sequence_examples.append(sequence_example)
    return (loaded_module_spec.name, new_sequence_examples, new_reward_stats)


def main(_):
  gin.parse_config_files_and_bindings(
      _GIN_FILES.value, bindings=_GIN_BINDINGS.value, skip_unknown=False)
  logging.info(gin.config_str())
  generate_trace()


def generate_trace(
    worker_manager_class=local_worker_manager.LocalWorkerPoolManager):

  config = registry.get_configuration()

  logging.info('Loading module specs from corpus.')
  module_filter = re.compile(
      _MODULE_FILTER.value) if _MODULE_FILTER.value else None

  cps = corpus.Corpus(
      data_path=_DATA_PATH.value,
      module_filter=lambda name: True
      if not module_filter else module_filter.match(name),
      additional_flags=config.flags_to_add(),
      delete_flags=config.flags_to_delete(),
      replace_flags=config.flags_to_replace())
  logging.info('Done loading module specs from corpus.')

  # Sampling if needed.
  sampled_modules = int(len(cps) * _SAMPLING_RATE.value)
  # sort files by size, to process the large files upfront, hopefully while
  # other smaller files are processed in parallel
  corpus_elements = cps.sample(k=sampled_modules, sort=True)

  tfrecord_context = (
      tf.io.TFRecordWriter(_OUTPUT_PATH.value)
      if _OUTPUT_PATH.value else contextlib.nullcontext())
  performance_context = (
      tf.io.gfile.GFile(_OUTPUT_PERFORMANCE_PATH.value, 'w')
      if _OUTPUT_PERFORMANCE_PATH.value else contextlib.nullcontext())
  work = [
      cps.load_module_spec(corpus_element) for corpus_element in corpus_elements
  ]

  runner_type = config.get_runner_type()
  with tfrecord_context as tfrecord_writer:
    with performance_context as performance_writer:
      with worker_manager_class(
          FilteringWorker,
          _NUM_WORKERS.value,
          policy_path=_POLICY_PATH.value,
          key_filter=_KEY_FILTER.value,
          runner_type=runner_type,
          runner_kwargs=worker.get_full_worker_args(
              runner_type, moving_average_decay_rate=0)) as lwm:

        _, result_futures = buffered_scheduler.schedule_on_worker_pool(
            action=lambda w, j: w.compile_and_filter(j),
            jobs=work,
            worker_pool=lwm)
        total_successful_examples = 0
        total_work = len(corpus_elements)
        total_failed_examples = 0
        total_training_examples = 0
        not_done = result_futures
        while not_done:
          (done, not_done) = concurrent.futures.wait(not_done, 10)
          succeeded = [
              r for r in done if not r.cancelled() and r.exception() is None
          ]
          total_successful_examples += len(succeeded)
          total_failed_examples += (len(done) - len(succeeded))
          for r in succeeded:
            module_name, records, reward_stat = r.result()
            if tfrecord_writer:
              total_training_examples += len(records)
              for r in records:
                tfrecord_writer.write(r)
            if performance_writer:
              for key, value in reward_stat.items():
                performance_writer.write(
                    (f'{module_name},{key},{value.default_reward},'
                     f'{value.moving_average_reward}\n'))
          logging.info('%d success, %d failed out of %d',
                       total_successful_examples, total_failed_examples,
                       total_work)

  print((f'{total_successful_examples} of {len(corpus_elements)} modules '
         f'succeeded, and {total_training_examples} trainining examples '
         'written'))


if __name__ == '__main__':
  flags.mark_flag_as_required('data_path')
  multiprocessing.handle_main(functools.partial(app.run, main))



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Train behavioral cloning policy."""

import os

from absl import app
from absl import flags
from absl import logging
import gin

# <Internal> Using XM - flags.  # pylint: disable=unused-import
from compiler_opt.rl import agent_config
from compiler_opt.rl import data_reader
from compiler_opt.rl import gin_external_configurables  # pylint: disable=unused-import
from compiler_opt.rl import policy_saver
from compiler_opt.rl import registry
from compiler_opt.rl import trainer

from tf_agents.agents import tf_agent
from tf_agents.policies import tf_policy

from typing import Dict

_ROOT_DIR = flags.DEFINE_string(
    'root_dir', os.getenv('TEST_UNDECLARED_OUTPUTS_DIR'),
    'Root directory for writing logs/summaries/checkpoints.')
_DATA_PATH = flags.DEFINE_multi_string(
    'data_path', [],
    'Path to TFRecord file(s) containing training data. Skip training and dump'
    'an untrained model with random weights (for testing purpose) if '
    'unspecified.')
_GIN_FILES = flags.DEFINE_multi_string(
    'gin_files', [], 'List of paths to gin configuration files.')
_GIN_BINDINGS = flags.DEFINE_multi_string(
    'gin_bindings', [],
    'Gin bindings to override the values set in the config files.')


@gin.configurable
def train_eval(agent_config_type=agent_config.BCAgentConfig,
               num_iterations=100,
               batch_size=64,
               train_sequence_length=1):
  """Train Behavioral Cloning."""
  root_dir = os.path.expanduser(_ROOT_DIR.value)
  root_dir = os.path.normpath(root_dir)
  problem_config = registry.get_configuration()
  time_step_spec, action_spec = problem_config.get_signature_spec()
  preprocessing_layer_creator = problem_config.get_preprocessing_layer_creator()

  # Initialize trainer and policy saver.
  agent_cfg: agent_config.AgentConfig = agent_config_type(
      time_step_spec=time_step_spec, action_spec=action_spec)
  agent: tf_policy.TFAgent = agent_config.create_agent(
      agent_cfg, preprocessing_layer_creator=preprocessing_layer_creator)
  llvm_trainer = trainer.Trainer(root_dir=root_dir, agent=agent)
  policy_dict: Dict[str, tf_policy.TFPolicy] = {
      'saved_policy': agent.policy,
      'saved_collect_policy': agent.collect_policy,
  }
  saver = policy_saver.PolicySaver(policy_dict=policy_dict)

  tfrecord_dataset_fn = data_reader.create_tfrecord_dataset_fn(
      agent_cfg=agent_cfg,
      batch_size=batch_size,
      train_sequence_length=train_sequence_length)

  # Train.
  if _DATA_PATH.value:
    dataset_iter = iter(tfrecord_dataset_fn(_DATA_PATH.value).repeat())
    monitor_dict = {}
    llvm_trainer.train(dataset_iter, monitor_dict, num_iterations)

  # Save final policy.
  saver.save(root_dir)


def main(_):
  gin.parse_config_files_and_bindings(
      _GIN_FILES.value, bindings=_GIN_BINDINGS.value, skip_unknown=False)
  logging.info(gin.config_str())

  train_eval()


if __name__ == '__main__':
  app.run(main)
